import os
from itertools import product

import numpy as np
from torch.functional import Tensor


def pwd():
    return os.path.dirname(os.path.realpath(__file__))


def heatmap2points(x: Tensor) -> Tensor:
    """B x C x H x W -> B x C x 2"""
    out = torch.zeros((x.size(0), x.size(1), 2))
    conf = torch.zeros((x.size(0), x.size(1), 1))
    for b, c in product(range(x.size(0)), range(x.size(1))):
        conf[b, c] = torch.max(x[b, c])
        out[b, c] = (x[b, c] == conf[b, c]).nonzero(as_tuple=False)[0]
        out[b, c] /= torch.tensor([x.size(2), x.size(3)])
    return out, conf


# draw 2d gaussians
def draw_labelmap(img: np.ndarray, pt, sigma=3, type="Gaussian"):
    # Draw a 2D gaussian
    # Adopted from https://github.com/anewell/pose-hg-train/blob/master/src/pypose/draw.py
    pt[0], pt[1] = pt[1], pt[0]

    # Check that any part of the gaussian is in-bounds
    ul = [int(pt[0] - 3 * sigma), int(pt[1] - 3 * sigma)]
    br = [int(pt[0] + 3 * sigma + 1), int(pt[1] + 3 * sigma + 1)]

    # Generate gaussian
    size = 6 * sigma + 1
    x = np.arange(0, size, 1, float)
    y = x[:, np.newaxis]
    x0 = y0 = size // 2
    # The gaussian is not normalized, we want the center value to equal 1
    if type == "Gaussian":
        g = np.exp(-((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))

    # Usable gaussian range
    g_x = max(0, -ul[0]), min(br[0], img.shape[1]) - ul[0]
    g_y = max(0, -ul[1]), min(br[1], img.shape[0]) - ul[1]
    # Image range
    img_x = max(0, ul[0]), min(br[0], img.shape[1])
    img_y = max(0, ul[1]), min(br[1], img.shape[0])

    img[img_y[0] : img_y[1], img_x[0] : img_x[1]] = g[g_y[0] : g_y[1], g_x[0] : g_x[1]]
    return img


import torch
from torch.functional import Tensor
from torchvision import utils


def tensorboard_plot_image(
    writer, x: Tensor, name: str, epoch: int, scale_each: bool = False
) -> None:
    """
    records single image into tensorboard.
    tensorboard logs are saved under lightning_logs.
    >>> SU.stensorboard_plot_image(
        self.logger.experiment,
        torch.cat([img1[:, :, -1], img2[:, :, -1]], dim=2),
        "recons",
        self.current_epoch,
    )
    """
    with torch.no_grad():
        x = x[:16].clone()
        grid = utils.make_grid(x, nrow=4, scale_each=scale_each, normalize=True)
        writer.add_image(name, grid, epoch)



import argparse
import os


def add_arguments(parser):
    parser.add_argument(
        "-j",
        "--workers",
        default=8,
        type=int,
        metavar="N",
        help="number of data loading workers (default: 4)",
    )
    parser.add_argument(
        "--epochs",
        default=200,
        type=int,
        metavar="N",
        help="number of total epochs to run",
    )
    # hyper-parameters
    parser.add_argument(
        "--train-batch",
        default=6,
        type=int,
        metavar="N",
        help="train batchsize",
        dest="train_batch",
    )
    parser.add_argument(
        "--test-batch",
        default=96,
        type=int,
        metavar="N",
        dest="test_batch",
        help="test batchsize",
    )
    parser.add_argument(
        "--lr",
        "--learning-rate",
        default=2.5e-4,
        type=float,
        metavar="LR",
        help="initial learning rate",
    )
    parser.add_argument(
        "--momentum", default=0, type=float, metavar="M", help="momentum"
    )
    parser.add_argument(
        "--weight-decay",
        "--wd",
        default=0,
        type=float,
        metavar="W",
        help="weight decay (default: 0)",
    )
    parser.add_argument(
        "--schedule",
        type=int,
        nargs="+",
        default=[25, 40, 70],
        help="Decrease learning rate at these epochs.",
    )
    parser.add_argument(
        "--gamma",
        type=float,
        default=0.1,
        help="LR is multiplied by gamma on schedule.",
    )

    parser.add_argument(
        "--num_classes", type=int, default=19,
    )

    parser.add_argument(
        "-s",
        "--stacks",
        default=2,
        type=int,
        metavar="N",
        help="Number of hourglasses to stack",
    )
    parser.add_argument(
        "--features",
        default=128,
        type=int,
        metavar="N",
        help="Number of features in the hourglass",
    )
    parser.add_argument(
        "-b",
        "--blocks",
        default=1,
        type=int,
        metavar="N",
        help="Number of residual modules at each location in the hourglass",
    )
    parser.add_argument("--inplanes", default=64, type=int, metavar="N")
    parser.add_argument("--stride", default=2, type=int, metavar="N")
    return parser


def create_parser():
    parser = argparse.ArgumentParser(description="DF3D Training")
    return add_arguments(parser)



from typing import *

import matplotlib.pyplot as plt
import numpy as np
import torch
from skimage.color import gray2rgb
from skimage.transform import resize

from df2d.util import draw_labelmap


class Drosophila2Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        inp: List[List[Union[str, np.ndarray]]],
        load_f: Callable = lambda x: plt.imread(x),
        img_size: List[int] = [256, 512],
        heatmap_size: List[int] = [64, 128],
        heatmap_channels: int = 19,
        return_heatmap: bool = False,
    ):
        self.inp = inp
        self.load_f = load_f
        self.img_size = img_size
        self.heatmap_size = heatmap_size
        self.return_heatmap = return_heatmap
        self.heatmap_channels = heatmap_channels

    def __getitem__(self, idx):
        img_path, pts2d = self.inp[idx]
        img = self.load_f(img_path)
        img = resize(img, self.img_size)

        # gray2rgb
        if img.ndim == 2:
            img = gray2rgb(img)

        # h w c -> c h w
        img = img.transpose(2, 0, 1)

        # remove the mean
        img = img - 0.22

        # draw heatmap
        if self.return_heatmap:
            hm = np.zeros(
                (self.heatmap_channels, self.heatmap_size[0], self.heatmap_size[1])
            )
            present_joints = [
                idx
                for idx in range(self.heatmap_channels)
                if not np.all(pts2d[idx] == 0)
            ]
            for idx in present_joints:
                hm[idx] = draw_labelmap(
                    hm[idx], pts2d[idx]  * np.array(self.heatmap_size)
                )

        if self.return_heatmap:
            return img, pts2d, tuple(self.inp[idx]), hm
        else:
            return img, pts2d, tuple(self.inp[idx])

    def __len__(self):
        return len(self.inp)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nely-df2d",
    version="0.13",
    packages=["df2d"],
    author="Semih Gunel",
    author_email="gunelsemih@gmail.com",
    description="2D pose estimation pipeline for tethered Drosophila.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NeLy-EPFL/Drosophila2DPose",
)



import os
import re
import subprocess
from itertools import product
from typing import *

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from df2d.dataset import Drosophila2Dataset
from df2d.model import Drosophila2DPose
from df2d.parser import create_parser
from df2d.util import heatmap2points, pwd

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def download_weights(path):
    command = f"curl -L -o {path} https://www.dropbox.com/s/csgon8uojr3gdd9/sh8_front_j8.tar?dl=0"
    print("Downloading network weights.")
    os.makedirs(os.path.dirname(path))
    subprocess.run(command, shell=True)


def inference_folder(
    folder: str,
    load_f: Callable = lambda x: plt.imread(x),
    args: Optional[Dict] = {},
    return_heatmap: bool = False,
    return_confidence: bool = False,
    max_img_id: Optional[int] = None,
):
    """processes all the images under a folder.
        returns normalized coordinates in [0, 1].
    >>> from df2d.inference import inference_folder
    >>> points2d = inference_folder('/home/user/Desktop/DeepFly3D/data/test/')
    >>> points2d.shape
    >>>     (7, 16, 19, 2) # n_cameras, n_images, n_joints, 2
    """
    checkpoint_path = os.path.join(pwd(), "../weights/sh8_deepfly.tar")
    if not os.path.exists(checkpoint_path):
        download_weights(checkpoint_path)
    args_default = create_parser().parse_args("").__dict__
    args_default.update(args)

    model = Drosophila2DPose(checkpoint_path=checkpoint_path, **args_default).to(device)

    inp = path2inp(
        folder, max_img_id=max_img_id
    )  # extract list of images under the folder
    dat = DataLoader(Drosophila2Dataset(inp, load_f=load_f), batch_size=8)

    return inference(
        model, dat, return_heatmap=return_heatmap, return_confidence=return_confidence
    )


import pickle


def pr2inp(path: str) -> List[str]:
    """converts pose result file into inp format
    >>> pr2inp("/data/test/df3d/df3d_result.pkl")
    >>>     [("/data/test/0.jpg", [[0,0], [.5,.5]]), ("/data/test/1.jpg", [[0,0], [.5,.5]])]
    """
    img_list = list()
    points2d = pickle.load(open(path, "rb"))["points2d"]
    n_cameras, n_images = points2d.shape[0] - 1, points2d.shape[1] - 1

    for cid, imgid in product(range(n_cameras), range(n_images)):
        img_path = os.path.join(
            os.path.dirname(path),
            "../camera_{cid}_img_{imgid}.jpg".format(cid=cid, imgid=imgid),
        )

        # skip if pose is missing
        if np.all(points2d[cid, imgid] == 0):
            continue

        if points2d[cid, imgid, 0, 0] == 0:
            pts = points2d[cid, imgid][19:]
            pts[..., 1] = 1 - pts[..., 1]
        else:
            pts = points2d[cid, imgid][:19]

        img_list.append((img_path, pts))

    return img_list


def path2inp(path: str, max_img_id: Optional[int] = None) -> List[str]:
    """
    >>> path2inp("/data/test/")
    >>>     ["/data/test/0.jpg", "/data/test/1.jpg"]
    """
    pattern = re.compile(r'camera_\d*_img_\d*.(jpg|png)')
    img_list = [
        (os.path.join(path, p), np.zeros((19)))
        for p in os.listdir(path)
        if pattern.match(p)
    ]

    if max_img_id is not None:
        img_list = [img for img in img_list if parse_img_path(img[0])[1] <= max_img_id]

    return img_list


def inference(
    model: Drosophila2DPose,
    dataset: Drosophila2Dataset,
    return_heatmap: bool = False,
    return_confidence: bool = False,
) -> np.ndarray:
    res = list()
    res_conf = list()
    heatmap = list()
    for batch in tqdm(dataset):
        x, _, d = batch
        hm = model(x)
        points, conf = heatmap2points(hm)
        points = points.cpu().data.numpy()
        conf = conf.cpu().data.numpy()
        if return_heatmap:
            heatmap.append(hm.cpu().data.numpy())
        for idx in range(x.size(0)):
            path = d[0][idx]
            res.append([path, points[idx]])
            res_conf.append([path, conf[idx]])

    points2d = inp2np(res)
    conf = inp2np(res_conf)

    if not return_heatmap and not return_confidence:
        return points2d

    ret = [points2d]
    if return_heatmap:
        ret.append(np.concatenate(heatmap, axis=0))
    if return_confidence:
        ret.append(conf)

    return ret


def parse_img_path(name: str) -> Tuple[int, int]:
    """returns cid and img_id"""
    name = os.path.basename(name)
    match = re.match(r"camera_(\d+)_img_(\d+)", name.replace(".jpg", ""))
    if match is None:
        print(f'Cannot parse image {name}')
    return int(match[1]), int(match[2])
        


def inp2np(inp: List) -> np.ndarray:
    """converts a list representation into numpy array in format C x J x 2
    each list element is a tuple, where the first element is a path in the camera_{camid}_img_{img_id} format.
    second element is a numpy array
    """
    n_cameras = max([parse_img_path(p)[0] for (p, _) in inp]) + 1
    n_images = max([parse_img_path(p)[1] for (p, _) in inp]) + 1

    n_joints = inp[0][1].shape[0]
    n_dim = inp[0][1].shape[1]

    points2d = np.zeros((n_cameras, n_images + 1, n_joints, n_dim))

    for (path, pts) in inp:
        cid, imgid = parse_img_path(path)
        points2d[cid, imgid] = pts

    return points2d



from typing import *

import pytorch_lightning as pl
import torch
from torch import nn
from torch.nn import functional as F

from df2d.util import heatmap2points, tensorboard_plot_image

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class Drosophila2DPose(pl.LightningModule):
    def __init__(
        self,
        checkpoint_path: str,
        stacks,
        blocks,
        num_classes,
        features,
        inplanes,
        stride,
        **args
    ):
        super().__init__()

        self.model = hg(
            num_stacks=stacks,
            num_blocks=blocks,
            num_classes=num_classes,
            num_feats=features,
            inplanes=inplanes,
            init_stride=stride,
        )

        if checkpoint_path is not None:
            pretrained = {
                k.replace("module.", ""): v
                for (k, v) in torch.load(checkpoint_path, map_location=device)[
                    "state_dict"
                ].items()
            }
            _ = self.model.load_state_dict(pretrained, strict=False)

    def forward(self, x):
        return self.model(x.float().to(device))[-1]

    def mse(self, hm_hat, y):
        y_hat, _ = heatmap2points(hm_hat).cuda()
        mse = torch.norm(y_hat - y, dim=2)
        return mse.mean()

    def configure_optimizers(self):
        optimizer = torch.optim.RMSprop(
            self.parameters(),
            lr=1e-4,
            weight_decay=1e-5,
        )
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, verbose=True, patience=5
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "train/loss",
        }

    def loss(self, hm_hat, hm):
        # fmt: off
        # missing joint predictions are set to all zeros, which is the same of the gt heatmap
        #   therefore giving zero loss
        missing_joints = torch.all(torch.flatten(hm, 2) == 0, dim=2, keepdim=True)
        missing_joints = missing_joints.float()
        missing_joints = missing_joints.unsqueeze(-1)
        hm_hat = hm_hat * (1 - missing_joints)
        # fmt: on
        return F.mse_loss(hm, hm_hat)

    def training_step(self, train_batch, batch_idx):
        x, d, y, hm = train_batch
        x = x.to(device).float()
        hm = hm.to(device).float()
        hm_hat = self.model(x)

        loss = self.loss(hm_hat[0], hm)
        for idx in range(1, len(hm_hat)):
            loss += self.loss(hm_hat[idx], hm)
        self.log("train/loss", loss, on_step=True)

        # log images
        if batch_idx == 0:
            tensorboard_plot_image(
                self.logger.experiment,
                torch.cat([x[:, [0], ::4, ::4], hm.sum(dim=1, keepdims=True)], dim=2),
                name="data",
                epoch=self.current_epoch,
                scale_each=True,
            )

            # log mse loss
            self.log(name="train/mse", value=self.mse(hm_hat[-1], d), prog_bar=True)

        return loss

    def validation_step(self, val_batch, batch_idx):
        x, d, y, hm = val_batch
        x = x.to(device).float()
        hm = hm.to(device).float()
        hm_hat = self.model(x.float())

        loss = self.loss(hm_hat[0], hm)
        for idx in range(1, len(hm_hat)):
            loss += self.loss(hm_hat[idx], hm)
        self.log("val/loss", loss, on_step=True)

        # log mse loss
        self.log(name="val/mse", value=self.mse(hm_hat[-1], d), prog_bar=True)

        return loss


"""
Hourglass network inserted in the pre-activated Resnet
"""
import torch.nn as nn
import torch.nn.functional as F


class Bottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()

        self.bn1 = nn.BatchNorm2d(inplanes)
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=True)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=stride, padding=1, bias=True
        )
        self.bn3 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 2, kernel_size=1, bias=True)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)

        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)

        out = self.bn3(out)
        out = self.relu(out)
        out = self.conv3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual

        return out


class Hourglass(nn.Module):
    def __init__(self, block, num_blocks, planes, depth):
        super(Hourglass, self).__init__()
        self.depth = depth
        self.block = block
        self.upsample = nn.Upsample(scale_factor=2)
        self.hg = self._make_hour_glass(block, num_blocks, planes, depth)

    def _make_residual(self, block, num_blocks, planes):
        layers = []
        for i in range(0, num_blocks):
            layers.append(block(planes * block.expansion, planes))
        return nn.Sequential(*layers)

    def _make_hour_glass(self, block, num_blocks, planes, depth):
        hg = []
        for i in range(depth):
            res = []
            for j in range(3):
                res.append(self._make_residual(block, num_blocks, planes))
            if i == 0:
                res.append(self._make_residual(block, num_blocks, planes))
            hg.append(nn.ModuleList(res))
        return nn.ModuleList(hg)

    def _hour_glass_forward(self, n, x):
        up1 = self.hg[n - 1][0](x)
        low1 = F.max_pool2d(x, 2, stride=2)
        low1 = self.hg[n - 1][1](low1)

        if n > 1:
            low2 = self._hour_glass_forward(n - 1, low1)
        else:
            low2 = self.hg[n - 1][3](low1)
        low3 = self.hg[n - 1][2](low2)
        up2 = self.upsample(low3)
        out = up1 + up2
        return out

    def forward(self, x):
        return self._hour_glass_forward(self.depth, x)


class HourglassNet(nn.Module):
    """Hourglass model from Newell et al ECCV 2016"""

    def __init__(
        self,
        block,
        num_stacks=2,
        num_blocks=4,
        num_classes=16,
        inplanes=64,
        num_feats=128,
        init_stride=2,
    ):
        super(HourglassNet, self).__init__()

        self.inplanes = inplanes
        self.num_feats = num_feats
        self.num_stacks = num_stacks
        self.conv1 = nn.Conv2d(
            3, self.inplanes, kernel_size=7, stride=init_stride, padding=3, bias=True
        )
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_residual(block, self.inplanes, 1)
        self.layer2 = self._make_residual(block, self.inplanes, 1)
        self.layer3 = self._make_residual(block, self.num_feats, 1)
        self.maxpool = nn.MaxPool2d(2, stride=2)

        # build hourglass modules
        ch = self.num_feats * block.expansion
        hg, res, fc, score, fc_, score_ = [], [], [], [], [], []
        for i in range(num_stacks):
            hg.append(Hourglass(block, num_blocks, self.num_feats, 4))
            res.append(self._make_residual(block, self.num_feats, num_blocks))
            fc.append(self._make_fc(ch, ch))
            score.append(nn.Conv2d(ch, num_classes, kernel_size=1, bias=True))
            if i < num_stacks - 1:
                fc_.append(nn.Conv2d(ch, ch, kernel_size=1, bias=True))
                score_.append(nn.Conv2d(num_classes, ch, kernel_size=1, bias=True))
        self.hg = nn.ModuleList(hg)
        self.res = nn.ModuleList(res)
        self.fc = nn.ModuleList(fc)
        self.score = nn.ModuleList(score)
        self.fc_ = nn.ModuleList(fc_)
        self.score_ = nn.ModuleList(score_)

    def _make_residual(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=True,
                ),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def _make_fc(self, inplanes, outplanes):
        bn = nn.BatchNorm2d(inplanes)
        conv = nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=True)
        return nn.Sequential(
            conv,
            bn,
            self.relu,
        )

    def forward(self, x):
        out = []
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.maxpool(x)
        x = self.layer2(x)
        x = self.layer3(x)

        for i in range(self.num_stacks):
            y = self.hg[i](x)
            y = self.res[i](y)
            y = self.fc[i](y)
            score = self.score[i](y)
            out.append(score)
            if i < self.num_stacks - 1:
                fc_ = self.fc_[i](y)
                score_ = self.score_[i](score)
                x = x + fc_ + score_

        return out


def hg(**kwargs):
    model = HourglassNet(
        Bottleneck,
        num_stacks=kwargs["num_stacks"],
        num_blocks=kwargs["num_blocks"],
        num_classes=kwargs["num_classes"],
        num_feats=kwargs["num_feats"],
        inplanes=kwargs["inplanes"],
        init_stride=kwargs["init_stride"],
    )
    return model
